from s2clientprotocol import sc2api_pb2 as sc_pb
from .protocol import Protocol
from .player import Computer

import logging
logger = logging.getLogger(__name__)

class Controller(Protocol):
    def __init__(self, ws, process):
        super().__init__(ws)
        self.__process = process

    @property
    def running(self):
        return self.__process._process is not None

    async def create_game(self, game_map, players, realtime, random_seed=None):
        assert isinstance(realtime, bool)
        req = sc_pb.RequestCreateGame(
            local_map=sc_pb.LocalMap(
                map_path=str(game_map.relative_path)
            ),
            realtime=realtime
        )
        if random_seed is not None:
            req.random_seed = random_seed

        for player in players:
            p = req.player_setup.add()
            p.type = player.type.value
            if isinstance(player, Computer):
                p.race = player.race.value
                p.difficulty = player.difficulty.value

        logger.info("Creating new game")
        logger.info(f"Map:     {game_map.name}")
        logger.info(f"Players: {', '.join(str(p) for p in players)}")
        result = await self._execute(create_game=req)
        return result



import random

from .unit import Unit
from .ids.unit_typeid import UnitTypeId
from .position import Point2, Point3
from typing import List, Dict, Set, Tuple, Any, Optional, Union  # mypy type checking


class Units(list):
    """A collection for units. Makes it easy to select units by selectors."""

    @classmethod
    def from_proto(cls, units, game_data):
        return cls((Unit(u, game_data) for u in units), game_data)

    def __init__(self, units, game_data):
        super().__init__(units)
        self.game_data = game_data

    def __call__(self, *args, **kwargs):
        return UnitSelection(self, *args, **kwargs)

    def select(self, *args, **kwargs):
        return UnitSelection(self, *args, **kwargs)

    def copy(self):
        return self.subgroup(self)

    def __or__(self, other: "Units") -> "Units":
        if self is None:
            return other
        if other is None:
            return self
        tags = {unit.tag for unit in self}
        units = self + [unit for unit in other if unit.tag not in tags]
        return Units(units, self.game_data)

    def __and__(self, other: "Units") -> "Units":
        if self is None:
            return other
        if other is None:
            return self
        tags = {unit.tag for unit in self}
        units = [unit for unit in other if unit.tag in tags]
        return Units(units, self.game_data)

    def __sub__(self, other: "Units") -> "Units":
        if self is None:
            return Units([], self.game_data)
        if other is None:
            return self
        tags = {unit.tag for unit in other}
        units = [unit for unit in self if unit.tag not in tags]
        return Units(units, self.game_data)

    def __hash__(self):
        return hash(unit.tag for unit in self)

    @property
    def amount(self) -> int:
        return len(self)

    @property
    def empty(self) -> bool:
        return not bool(self)

    @property
    def exists(self) -> bool:
        return bool(self)

    def find_by_tag(self, tag) -> Optional[Unit]:
        for unit in self:
            if unit.tag == tag:
                return unit
        return None

    def by_tag(self, tag):
        unit = self.find_by_tag(tag)
        if unit is None:
            raise KeyError("Unit not found")
        return unit

    @property
    def first(self) -> Unit:
        assert self
        return self[0]

    def take(self, n: int, require_all: bool = True) -> "Units":
        assert (not require_all) or len(self) >= n
        return self[:n]

    @property
    def random(self) -> Unit:
        assert self.exists
        return random.choice(self)

    def random_or(self, other: any) -> Unit:
        if self.exists:
            return random.choice(self)
        else:
            return other

    def random_group_of(self, n):
        # TODO allow n > amount with n = min(n,amount)?
        assert 0 <= n <= self.amount
        if n == 0:
            return self.subgroup([])
        elif self.amount == n:
            return self
        else:
            return self.subgroup(random.sample(self, n))

    def in_attack_range_of(self, unit: Unit, bonus_distance: Union[int, float] = 0) -> "Units":
        """ Filters units that are in attack range of the unit in parameter """
        return self.filter(lambda x: unit.target_in_range(x, bonus_distance=bonus_distance))

    def closest_distance_to(self, position: Union[Unit, Point2, Point3]) -> Union[int, float]:
        """ Returns the distance between the closest unit from this group to the target unit """
        assert self.exists
        if isinstance(position, Unit):
            position = position.position
        return position.distance_to_closest(
            [u.position for u in self]
        )  # Note: list comprehension creation is 0-5% faster than set comprehension

    def furthest_distance_to(self, position: Union[Unit, Point2, Point3]) -> Union[int, float]:
        """ Returns the distance between the furthest unit from this group to the target unit """
        assert self.exists
        if isinstance(position, Unit):
            position = position.position
        return position.distance_to_furthest([u.position for u in self])

    def closest_to(self, position: Union[Unit, Point2, Point3]) -> Unit:
        assert self.exists
        if isinstance(position, Unit):
            position = position.position
        return position.closest(self)

    def furthest_to(self, position: Union[Unit, Point2, Point3]) -> Unit:
        assert self.exists
        if isinstance(position, Unit):
            position = position.position
        return position.furthest(self)

    def closer_than(self, distance: Union[int, float], position: Union[Unit, Point2, Point3]) -> "Units":
        if isinstance(position, Unit):
            position = position.position
        distance_squared = distance ** 2
        return self.filter(lambda unit: unit.position._distance_squared(position.to2) < distance_squared)

    def further_than(self, distance: Union[int, float], position: Union[Unit, Point2, Point3]) -> "Units":
        if isinstance(position, Unit):
            position = position.position
        distance_squared = distance ** 2
        return self.filter(lambda unit: unit.position._distance_squared(position.to2) > distance_squared)

    def subgroup(self, units):
        return Units(list(units), self.game_data)

    def filter(self, pred: callable) -> "Units":
        return self.subgroup(filter(pred, self))

    def sorted(self, keyfn: callable, reverse: bool = False) -> "Units":
        if len(self) in {0, 1}:
            return self
        return self.subgroup(sorted(self, key=keyfn, reverse=reverse))

    def sorted_by_distance_to(self, position: Union[Unit, Point2], reverse: bool = False) -> "Units":
        """ This function should be a bit faster than using units.sorted(keyfn=lambda u: u.distance_to(position)) """
        if len(self) in [0, 1]:
            return self
        position = position.position
        return self.sorted(keyfn=lambda unit: unit.position._distance_squared(position), reverse=reverse)

    def tags_in(self, other: Union[Set[int], List[int], Dict[int, Any]]) -> "Units":
        """ Filters all units that have their tags in the 'other' set/list/dict """
        # example: self.units(QUEEN).tags_in(self.queen_tags_assigned_to_do_injects)
        if isinstance(other, list):
            other = set(other)
        return self.filter(lambda unit: unit.tag in other)

    def tags_not_in(self, other: Union[Set[int], List[int], Dict[int, Any]]) -> "Units":
        """ Filters all units that have their tags not in the 'other' set/list/dict """
        # example: self.units(QUEEN).tags_not_in(self.queen_tags_assigned_to_do_injects)
        if isinstance(other, list):
            other = set(other)
        return self.filter(lambda unit: unit.tag not in other)

    def of_type(self, other: Union[UnitTypeId, Set[UnitTypeId], List[UnitTypeId], Dict[UnitTypeId, Any]]) -> "Units":
        """ Filters all units that are of a specific type """
        # example: self.units.of_type([ZERGLING, ROACH, HYDRALISK, BROODLORD])
        if isinstance(other, UnitTypeId):
            other = {other}
        if isinstance(other, list):
            other = set(other)
        return self.filter(lambda unit: unit.type_id in other)

    def exclude_type(
        self, other: Union[UnitTypeId, Set[UnitTypeId], List[UnitTypeId], Dict[UnitTypeId, Any]]
    ) -> "Units":
        """ Filters all units that are not of a specific type """
        # example: self.known_enemy_units.exclude_type([OVERLORD])
        if isinstance(other, UnitTypeId):
            other = {other}
        if isinstance(other, list):
            other = set(other)
        return self.filter(lambda unit: unit.type_id not in other)

    def same_tech(self, other: Union[UnitTypeId, Set[UnitTypeId], List[UnitTypeId], Dict[UnitTypeId, Any]]) -> "Units":
        """ Usage:
        'self.units.same_tech(UnitTypeId.COMMANDCENTER)' or 'self.units.same_tech(UnitTypeId.ORBITALCOMMAND)'
        returns all CommandCenter, CommandCenterFlying, OrbitalCommand, OrbitalCommandFlying, PlanetaryFortress
        This also works with a set/list/dict parameter, e.g. 'self.units.same_tech({UnitTypeId.COMMANDCENTER, UnitTypeId.SUPPLYDEPOT})'
        Untested: This should return the equivalents for Hatchery, WarpPrism, Observer, Overseer, SupplyDepot and others
        """
        if isinstance(other, UnitTypeId):
            other = {other}
        tech_alias_types = set(other)
        for unitType in other:
            tech_alias = self.game_data.units[unitType.value].tech_alias
            if tech_alias:
                for same in tech_alias:
                    tech_alias_types.add(same)
        return self.filter(
            lambda unit: unit.type_id in tech_alias_types
            or unit._type_data.tech_alias is not None
            and any(same in tech_alias_types for same in unit._type_data.tech_alias)
        )

    def same_unit(self, other: Union[UnitTypeId, Set[UnitTypeId], List[UnitTypeId], Dict[UnitTypeId, Any]]) -> "Units":
        """ Usage:
        'self.units.same_tech(UnitTypeId.COMMANDCENTER)'
        returns CommandCenter and CommandCenterFlying,
        'self.units.same_tech(UnitTypeId.ORBITALCOMMAND)'
        returns OrbitalCommand and OrbitalCommandFlying
        This also works with a set/list/dict parameter, e.g. 'self.units.same_tech({UnitTypeId.COMMANDCENTER, UnitTypeId.SUPPLYDEPOT})'
        Untested: This should return the equivalents for WarpPrism, Observer, Overseer, SupplyDepot and others
        """
        if isinstance(other, UnitTypeId):
            other = {other}
        unit_alias_types = set(other)
        for unitType in other:
            unit_alias = self.game_data.units[unitType.value].unit_alias
            if unit_alias:
                unit_alias_types.add(unit_alias)
        return self.filter(
            lambda unit: unit.type_id in unit_alias_types
            or unit._type_data.unit_alias is not None
            and unit._type_data.unit_alias in unit_alias_types
        )

    @property
    def center(self) -> Point2:
        """ Returns the central point of all units in this list """
        assert self
        pos = Point2(
            (
                sum([unit.position.x for unit in self]) / self.amount,
                sum([unit.position.y for unit in self]) / self.amount,
            )
        )
        return pos

    @property
    def selected(self) -> "Units":
        return self.filter(lambda unit: unit.is_selected)

    @property
    def tags(self) -> Set[int]:
        return {unit.tag for unit in self}

    @property
    def ready(self) -> "Units":
        return self.filter(lambda unit: unit.is_ready)

    @property
    def not_ready(self) -> "Units":
        return self.filter(lambda unit: not unit.is_ready)

    @property
    def noqueue(self) -> "Units":
        return self.filter(lambda unit: unit.noqueue)

    @property
    def idle(self) -> "Units":
        return self.filter(lambda unit: unit.is_idle)

    @property
    def owned(self) -> "Units":
        return self.filter(lambda unit: unit.is_mine)

    @property
    def enemy(self) -> "Units":
        return self.filter(lambda unit: unit.is_enemy)

    @property
    def flying(self) -> "Units":
        return self.filter(lambda unit: unit.is_flying)

    @property
    def not_flying(self) -> "Units":
        return self.filter(lambda unit: not unit.is_flying)

    @property
    def structure(self) -> "Units":
        return self.filter(lambda unit: unit.is_structure)

    @property
    def not_structure(self) -> "Units":
        return self.filter(lambda unit: not unit.is_structure)

    @property
    def gathering(self) -> "Units":
        return self.filter(lambda unit: unit.is_gathering)

    @property
    def returning(self) -> "Units":
        return self.filter(lambda unit: unit.is_returning)

    @property
    def collecting(self) -> "Units":
        return self.filter(lambda unit: unit.is_collecting)

    @property
    def visible(self) -> "Units":
        return self.filter(lambda unit: unit.is_visible)

    @property
    def mineral_field(self) -> "Units":
        return self.filter(lambda unit: unit.is_mineral_field)

    @property
    def vespene_geyser(self) -> "Units":
        return self.filter(lambda unit: unit.is_vespene_geyser)

    @property
    def prefer_idle(self) -> "Units":
        return self.sorted(lambda unit: unit.is_idle, reverse=True)

    def prefer_close_to(self, p: Union[Unit, Point2, Point3]) -> "Units":
        # TODO redundant?
        return self.sorted_by_distance_to(p)


class UnitSelection(Units):
    def __init__(self, parent, unit_type_id=None):
        assert unit_type_id is None or isinstance(unit_type_id, (UnitTypeId, set))
        if isinstance(unit_type_id, set):
            assert all(isinstance(t, UnitTypeId) for t in unit_type_id)

        self.unit_type_id = unit_type_id
        super().__init__([u for u in parent if self.matches(u)], parent.game_data)

    def matches(self, unit):
        if self.unit_type_id is None:
            # empty selector matches everything
            return True
        elif isinstance(self.unit_type_id, set):
            return unit.type_id in self.unit_type_id
        else:
            return self.unit_type_id == unit.type_id



# DO NOT EDIT!
# This file was automatically generated by "generate_id_constants_from_stableid.py"

import enum

class UpgradeId(enum.Enum):
    NULL = 0
    CARRIERLAUNCHSPEEDUPGRADE = 1
    GLIALRECONSTITUTION = 2
    TUNNELINGCLAWS = 3
    CHITINOUSPLATING = 4
    HISECAUTOTRACKING = 5
    TERRANBUILDINGARMOR = 6
    TERRANINFANTRYWEAPONSLEVEL1 = 7
    TERRANINFANTRYWEAPONSLEVEL2 = 8
    TERRANINFANTRYWEAPONSLEVEL3 = 9
    NEOSTEELFRAME = 10
    TERRANINFANTRYARMORSLEVEL1 = 11
    TERRANINFANTRYARMORSLEVEL2 = 12
    TERRANINFANTRYARMORSLEVEL3 = 13
    REAPERSPEED = 14
    STIMPACK = 15
    SHIELDWALL = 16
    PUNISHERGRENADES = 17
    SIEGETECH = 18
    HIGHCAPACITYBARRELS = 19
    BANSHEECLOAK = 20
    MEDIVACCADUCEUSREACTOR = 21
    RAVENCORVIDREACTOR = 22
    HUNTERSEEKER = 23
    DURABLEMATERIALS = 24
    PERSONALCLOAKING = 25
    GHOSTMOEBIUSREACTOR = 26
    TERRANVEHICLEARMORSLEVEL1 = 27
    TERRANVEHICLEARMORSLEVEL2 = 28
    TERRANVEHICLEARMORSLEVEL3 = 29
    TERRANVEHICLEWEAPONSLEVEL1 = 30
    TERRANVEHICLEWEAPONSLEVEL2 = 31
    TERRANVEHICLEWEAPONSLEVEL3 = 32
    TERRANSHIPARMORSLEVEL1 = 33
    TERRANSHIPARMORSLEVEL2 = 34
    TERRANSHIPARMORSLEVEL3 = 35
    TERRANSHIPWEAPONSLEVEL1 = 36
    TERRANSHIPWEAPONSLEVEL2 = 37
    TERRANSHIPWEAPONSLEVEL3 = 38
    PROTOSSGROUNDWEAPONSLEVEL1 = 39
    PROTOSSGROUNDWEAPONSLEVEL2 = 40
    PROTOSSGROUNDWEAPONSLEVEL3 = 41
    PROTOSSGROUNDARMORSLEVEL1 = 42
    PROTOSSGROUNDARMORSLEVEL2 = 43
    PROTOSSGROUNDARMORSLEVEL3 = 44
    PROTOSSSHIELDSLEVEL1 = 45
    PROTOSSSHIELDSLEVEL2 = 46
    PROTOSSSHIELDSLEVEL3 = 47
    OBSERVERGRAVITICBOOSTER = 48
    GRAVITICDRIVE = 49
    EXTENDEDTHERMALLANCE = 50
    HIGHTEMPLARKHAYDARINAMULET = 51
    PSISTORMTECH = 52
    ZERGMELEEWEAPONSLEVEL1 = 53
    ZERGMELEEWEAPONSLEVEL2 = 54
    ZERGMELEEWEAPONSLEVEL3 = 55
    ZERGGROUNDARMORSLEVEL1 = 56
    ZERGGROUNDARMORSLEVEL2 = 57
    ZERGGROUNDARMORSLEVEL3 = 58
    ZERGMISSILEWEAPONSLEVEL1 = 59
    ZERGMISSILEWEAPONSLEVEL2 = 60
    ZERGMISSILEWEAPONSLEVEL3 = 61
    OVERLORDSPEED = 62
    OVERLORDTRANSPORT = 63
    BURROW = 64
    ZERGLINGATTACKSPEED = 65
    ZERGLINGMOVEMENTSPEED = 66
    HYDRALISKSPEED = 67
    ZERGFLYERWEAPONSLEVEL1 = 68
    ZERGFLYERWEAPONSLEVEL2 = 69
    ZERGFLYERWEAPONSLEVEL3 = 70
    ZERGFLYERARMORSLEVEL1 = 71
    ZERGFLYERARMORSLEVEL2 = 72
    ZERGFLYERARMORSLEVEL3 = 73
    INFESTORENERGYUPGRADE = 74
    CENTRIFICALHOOKS = 75
    BATTLECRUISERENABLESPECIALIZATIONS = 76
    BATTLECRUISERBEHEMOTHREACTOR = 77
    PROTOSSAIRWEAPONSLEVEL1 = 78
    PROTOSSAIRWEAPONSLEVEL2 = 79
    PROTOSSAIRWEAPONSLEVEL3 = 80
    PROTOSSAIRARMORSLEVEL1 = 81
    PROTOSSAIRARMORSLEVEL2 = 82
    PROTOSSAIRARMORSLEVEL3 = 83
    WARPGATERESEARCH = 84
    HALTECH = 85
    CHARGE = 86
    BLINKTECH = 87
    ANABOLICSYNTHESIS = 88
    OBVERSEINCUBATION = 89
    VIKINGJOTUNBOOSTERS = 90
    ORGANICCARAPACE = 91
    INFESTORPERISTALSIS = 92
    ABDOMINALFORTITUDE = 93
    HYDRALISKSPEEDUPGRADE = 94
    BANELINGBURROWMOVE = 95
    COMBATDRUGS = 96
    STRIKECANNONS = 97
    TRANSFORMATIONSERVOS = 98
    PHOENIXRANGEUPGRADE = 99
    TEMPESTRANGEUPGRADE = 100
    NEURALPARASITE = 101
    LOCUSTLIFETIMEINCREASE = 102
    ULTRALISKBURROWCHARGEUPGRADE = 103
    ORACLEENERGYUPGRADE = 104
    RESTORESHIELDS = 105
    PROTOSSHEROSHIPWEAPON = 106
    PROTOSSHEROSHIPDETECTOR = 107
    PROTOSSHEROSHIPSPELL = 108
    REAPERJUMP = 109
    INCREASEDRANGE = 110
    ZERGBURROWMOVE = 111
    ANIONPULSECRYSTALS = 112
    TERRANVEHICLEANDSHIPWEAPONSLEVEL1 = 113
    TERRANVEHICLEANDSHIPWEAPONSLEVEL2 = 114
    TERRANVEHICLEANDSHIPWEAPONSLEVEL3 = 115
    TERRANVEHICLEANDSHIPARMORSLEVEL1 = 116
    TERRANVEHICLEANDSHIPARMORSLEVEL2 = 117
    TERRANVEHICLEANDSHIPARMORSLEVEL3 = 118
    FLYINGLOCUSTS = 119
    ROACHSUPPLY = 120
    IMMORTALREVIVE = 121
    DRILLCLAWS = 122
    CYCLONELOCKONRANGEUPGRADE = 123
    CYCLONEAIRUPGRADE = 124
    LIBERATORMORPH = 125
    ADEPTSHIELDUPGRADE = 126
    LURKERRANGE = 127
    IMMORTALBARRIER = 128
    ADEPTKILLBOUNCE = 129
    ADEPTPIERCINGATTACK = 130
    CINEMATICMODE = 131
    CURSORDEBUG = 132
    MAGFIELDLAUNCHERS = 133
    EVOLVEGROOVEDSPINES = 134
    EVOLVEMUSCULARAUGMENTS = 135
    BANSHEESPEED = 136
    MEDIVACRAPIDDEPLOYMENT = 137
    RAVENRECALIBRATEDEXPLOSIVES = 138
    MEDIVACINCREASESPEEDBOOST = 139
    LIBERATORAGRANGEUPGRADE = 140
    DARKTEMPLARBLINKUPGRADE = 141
    RAVAGERRANGE = 142
    RAVENDAMAGEUPGRADE = 143
    CYCLONELOCKONDAMAGEUPGRADE = 144
    ARESCLASSWEAPONSSYSTEMVIKING = 145
    AUTOHARVESTER = 146
    HYBRIDCPLASMAUPGRADEHARD = 147
    HYBRIDCPLASMAUPGRADEINSANE = 148
    INTERCEPTORLIMIT4 = 149
    INTERCEPTORLIMIT6 = 150
    _330MMBARRAGECANNONS = 151
    NOTPOSSIBLESIEGEMODE = 152
    NEOSTEELFRAME_2 = 153
    NEOSTEELANDSHRIKETURRETICONUPGRADE = 154
    OCULARIMPLANTS = 155
    CROSSSPECTRUMDAMPENERS = 156
    ORBITALSTRIKE = 157
    CLUSTERBOMB = 158
    SHAPEDHULL = 159
    SPECTRETOOLTIPUPGRADE = 160
    ULTRACAPACITORS = 161
    VANADIUMPLATING = 162
    COMMANDCENTERREACTOR = 163
    REGENERATIVEBIOSTEEL = 164
    CELLULARREACTORS = 165
    BANSHEECLOAKEDDAMAGE = 166
    DISTORTIONBLASTERS = 167
    EMPTOWER = 168
    SUPPLYDEPOTDROP = 169
    HIVEMINDEMULATOR = 170
    FORTIFIEDBUNKERCARAPACE = 171
    PREDATOR = 172
    SCIENCEVESSEL = 173
    DUALFUSIONWELDERS = 174
    ADVANCEDCONSTRUCTION = 175
    ADVANCEDMEDICTRAINING = 176
    PROJECTILEACCELERATORS = 177
    REINFORCEDSUPERSTRUCTURE = 178
    MULE = 179
    ORBITALRELAY = 180
    RAZORWIRE = 181
    ADVANCEDHEALINGAI = 182
    TWINLINKEDFLAMETHROWERS = 183
    NANOCONSTRUCTOR = 184
    CERBERUSMINES = 185
    HYPERFLUXOR = 186
    TRILITHIUMPOWERCELLS = 187
    PERMANENTCLOAKGHOST = 188
    PERMANENTCLOAKSPECTRE = 189
    ULTRASONICPULSE = 190
    SURVIVALPODS = 191
    ENERGYSTORAGE = 192
    FULLBORECANISTERAMMO = 193
    CAMPAIGNJOTUNBOOSTERS = 194
    MICROFILTERING = 195
    PARTICLECANNONAIR = 196
    VULTUREAUTOREPAIR = 197
    PSIDISRUPTOR = 198
    SCIENCEVESSELENERGYMANIPULATION = 199
    SCIENCEVESSELPLASMAWEAPONRY = 200
    SHOWGATLINGGUN = 201
    TECHREACTOR = 202
    TECHREACTORAI = 203
    TERRANDEFENSERANGEBONUS = 204
    X88TNAPALMUPGRADE = 205
    HURRICANEMISSILES = 206
    MECHANICALREBIRTH = 207
    MARINESTIMPACK = 208
    DARKTEMPLARTACTICS = 209
    CLUSTERWARHEADS = 210
    CLOAKDISTORTIONFIELD = 211
    DEVASTATORMISSILES = 212
    DISTORTIONTHRUSTERS = 213
    DYNAMICPOWERROUTING = 214
    IMPALERROUNDS = 215
    KINETICFIELDS = 216
    BURSTCAPACITORS = 217
    HAILSTORMMISSILEPODS = 218
    RAPIDDEPLOYMENT = 219
    REAPERSTIMPACK = 220
    REAPERD8CHARGE = 221
    TYCHUS05BATTLECRUISERPENETRATION = 222
    VIRALPLASMA = 223
    FIREBATJUGGERNAUTPLATING = 224
    MULTILOCKTARGETINGSYSTEMS = 225
    TURBOCHARGEDENGINES = 226
    DISTORTIONSENSORS = 227
    INFERNALPREIGNITERS = 228
    HELLIONCAMPAIGNINFERNALPREIGNITER = 229
    NAPALMFUELTANKS = 230
    AUXILIARYMEDBOTS = 231
    JUGGERNAUTPLATING = 232
    MARAUDERLIFEBOOST = 233
    COMBATSHIELD = 234
    REAPERU238ROUNDS = 235
    MAELSTROMROUNDS = 236
    SIEGETANKSHAPEDBLAST = 237
    TUNGSTENSPIKES = 238
    BEARCLAWNOZZLES = 239
    NANOBOTINJECTORS = 240
    STABILIZERMEDPACKS = 241
    HALOROCKETS = 242
    SCAVENGINGSYSTEMS = 243
    EXTRAMINES = 244
    ARESCLASSWEAPONSSYSTEM = 245
    WHITENAPALM = 246
    VIRALMUNITIONS = 247
    JACKHAMMERCONCUSSIONGRENADES = 248
    FIRESUPPRESSIONSYSTEMS = 249
    FLARERESEARCH = 250
    MODULARCONSTRUCTION = 251
    EXPANDEDHULL = 252
    SHRIKETURRET = 253
    MICROFUSIONREACTORS = 254
    WRAITHCLOAK = 255
    SINGULARITYCHARGE = 256
    GRAVITICTHRUSTERS = 257
    YAMATOCANNON = 258
    DEFENSIVEMATRIX = 259
    DARKPROTOSS = 260
    TERRANINFANTRYWEAPONSULTRACAPACITORSLEVEL1 = 261
    TERRANINFANTRYWEAPONSULTRACAPACITORSLEVEL2 = 262
    TERRANINFANTRYWEAPONSULTRACAPACITORSLEVEL3 = 263
    TERRANINFANTRYARMORSVANADIUMPLATINGLEVEL1 = 264
    TERRANINFANTRYARMORSVANADIUMPLATINGLEVEL2 = 265
    TERRANINFANTRYARMORSVANADIUMPLATINGLEVEL3 = 266
    TERRANVEHICLEWEAPONSULTRACAPACITORSLEVEL1 = 267
    TERRANVEHICLEWEAPONSULTRACAPACITORSLEVEL2 = 268
    TERRANVEHICLEWEAPONSULTRACAPACITORSLEVEL3 = 269
    TERRANVEHICLEARMORSVANADIUMPLATINGLEVEL1 = 270
    TERRANVEHICLEARMORSVANADIUMPLATINGLEVEL2 = 271
    TERRANVEHICLEARMORSVANADIUMPLATINGLEVEL3 = 272
    TERRANSHIPWEAPONSULTRACAPACITORSLEVEL1 = 273
    TERRANSHIPWEAPONSULTRACAPACITORSLEVEL2 = 274
    TERRANSHIPWEAPONSULTRACAPACITORSLEVEL3 = 275
    TERRANSHIPARMORSVANADIUMPLATINGLEVEL1 = 276
    TERRANSHIPARMORSVANADIUMPLATINGLEVEL2 = 277
    TERRANSHIPARMORSVANADIUMPLATINGLEVEL3 = 278
    HIREKELMORIANMINERSPH = 279
    HIREDEVILDOGSPH = 280
    HIRESPARTANCOMPANYPH = 281
    HIREHAMMERSECURITIESPH = 282
    HIRESIEGEBREAKERSPH = 283
    HIREHELSANGELSPH = 284
    HIREDUSKWINGPH = 285
    HIREDUKESREVENGE = 286
    TOSHEASYMODE = 287
    VOIDRAYSPEEDUPGRADE = 288
    SMARTSERVOS = 289
    ARMORPIERCINGROCKETS = 290
    CYCLONERAPIDFIRELAUNCHERS = 291
    RAVENENHANCEDMUNITIONS = 292
    DIGGINGCLAWS = 293
    CARRIERCARRIERCAPACITY = 294
    CARRIERLEASHRANGEUPGRADE = 295

for item in UpgradeId:
    assert not item.name in globals()
    globals()[item.name] = item



import os
from pathlib import Path
import platform
import re
import logging
logger = logging.getLogger(__name__)

BASEDIR = {
    "Windows": "C:/Program Files (x86)/StarCraft II",
    "Darwin": "/Applications/StarCraft II",
    "Linux": "~/StarCraftII",
    "WineLinux": "~/.wine/drive_c/Program Files (x86)/StarCraft II",
}

USERPATH = {
    "Windows": "\\Documents\\StarCraft II\\ExecuteInfo.txt",
    "Darwin": "/Library/Application Support/Blizzard/StarCraft II/ExecuteInfo.txt",
    "Linux": None,
    "WineLinux": None,
}

BINPATH = {
    "Windows": "SC2_x64.exe",
    "Darwin": "SC2.app/Contents/MacOS/SC2",
    "Linux": "SC2_x64",
    "WineLinux": "SC2_x64.exe",
}

CWD = {
    "Windows": "Support64",
    "Darwin": None,
    "Linux": None,
    "WineLinux": "Support64",
}

PF = os.environ.get("SC2PF", platform.system())

def get_env():
    # TODO: Linux env conf from: https://github.com/deepmind/pysc2/blob/master/pysc2/run_configs/platforms.py
    return None

def latest_executeble(versions_dir):
    latest = max((int(p.name[4:]), p) for p in versions_dir.iterdir() if p.is_dir() and p.name.startswith("Base"))
    version, path = latest
    if version < 55958:
        logger.critical(f"Your SC2 binary is too old. Upgrade to 3.16.1 or newer.")
        exit(1)
    return path / BINPATH[PF]


class _MetaPaths(type):
    """"Lazily loads paths to allow importing the library even if SC2 isn't installed."""
    def __setup(self):
        if PF not in BASEDIR:
            logger.critical(f"Unsupported platform '{PF}'")
            exit(1)

        try:
            base = os.environ.get("SC2PATH")
            if base is None and USERPATH[PF] is not None:
                einfo = str(Path.home().expanduser()) + USERPATH[PF]
                if os.path.isfile(einfo):
                    with open(einfo) as f:
                        content = f.read()
                    if content:
                        base = re.search(r" = (.*)Versions", content).group(1)
                        if not os.path.exists(base):
                            base = None
            if base is None:
                base = BASEDIR[PF]
            self.BASE = Path(base).expanduser()
            self.EXECUTABLE = latest_executeble(self.BASE / "Versions")
            self.CWD = self.BASE / CWD[PF] if CWD[PF] else None

            self.REPLAYS = self.BASE / "Replays"


            if (self.BASE / "maps").exists():
                self.MAPS = self.BASE / "maps"
            else:
                self.MAPS = self.BASE / "Maps"
        except FileNotFoundError as e:
            logger.critical(f"SC2 installation not found: File '{e.filename}' does not exist.")
            exit(1)

    def __getattr__(self, attr):
        self.__setup()
        return getattr(self, attr)

class Paths(metaclass=_MetaPaths):
    """Paths for SC2 folders, lazily loaded using the above metaclass."""



# This file contains temporary fixes for issues already solved in the
# blizzard_internal branch of s2client-proto.
# These should be removed as soon as the changes are live and implemented.
#
# https://github.com/Blizzard/s2client-proto/tree/blizzard_internal
#
# NO OTHER CONTENT IS ALLOWED HERE



import asyncio
import logging
from multiprocessing import Pipe
from select import select
from threading import Thread

import gym
from gym.spaces import Box
from sc2 import run_game, maps, Race
from sc2.player import Bot, AbstractPlayer

from .Sc2Bot import Sc2Bot

logger = logging.getLogger(__name__)

class Sc2Env(gym.Env):
    def __init__(self, map, bot_race, get_opponents,
            initializer, observer, actuator, **kwargs):
        self.map = maps.get(map)
        self.training_bot = Bot(Race[bot_race], EnvBot())
        self.get_opponents = get_opponents
        self.initializer = initializer
        self.observer = observer
        self.actuator = actuator
        self.kwargs = kwargs

        self.env_pipe, self.runner_pipe = None, None
        self.thread = None

        self.reset()

    def _run_game(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        players = [
            Bot(
                Race[bot_race], 
                Sc2Bot(self.runner_pipe, self.initializer, self.observer, self.actuator)),
            *self.get_opponents()]
        assert any(isinstance(player, AbstractPlayer) for player in players)

        run_game(self.map, players, **self.kwargs)

    def reset(self):
        if self.thread and self.env_pipe:
            self.env_pipe.send(None)
            self.thread.join()
        
        self.env_pipe, self.runner_pipe = Pipe(True)
        self.thread = Thread(target=self._run_game).start()
        
        logger.info("waiting Sc2Bot spaces")
        select([self.env_pipe], [], [], 60)
        self.action_space, self.observation_space = self.env_pipe.recv()

        logger.debug("waiting Sc2Bot observation")
        select([self.env_pipe], [], [], 60)
        self.observation, self.reward = self.env_pipe.recv()

        return self.observation

    def step(self, action):
        self.env_pipe.send(action)

        logger.debug("waiting Sc2Env observation")
        select([self.env_pipe], [], [])
        msg = self.env_pipe.recv()

        done = bool(msg)
        if not done:
            self.observation, self.reward =  msg

        return self.observation, self.reward, done, {}



from collections import deque
from typing import Any, Dict, FrozenSet, Generator, List, Optional, Sequence, Set, Tuple, Union

from .pixel_map import PixelMap
from .player import Player
from .position import Point2, Rect, Size


class Ramp:
    def __init__(self, points: Set[Point2], game_info: "GameInfo"):
        self._points: Set[Point2] = points
        self.__game_info = game_info
        # tested by printing actual building locations vs calculated depot positions
        self.x_offset = 0.5  # might be errors with the pixelmap?
        self.y_offset = -0.5

    @property
    def _height_map(self):
        return self.__game_info.terrain_height

    @property
    def _placement_grid(self):
        return self.__game_info.placement_grid

    @property
    def size(self) -> int:
        return len(self._points)

    def height_at(self, p: Point2) -> int:
        return self._height_map[p]

    @property
    def points(self) -> Set[Point2]:
        return self._points.copy()

    @property
    def upper(self) -> Set[Point2]:
        """ Returns the upper points of a ramp. """
        max_height = max([self.height_at(p) for p in self._points])
        return {p for p in self._points if self.height_at(p) == max_height}

    @property
    def upper2_for_ramp_wall(self) -> Set[Point2]:
        """ Returns the 2 upper ramp points of the main base ramp required for the supply depot and barracks placement properties used in this file. """
        if len(self.upper) > 5:
            # NOTE: this was way too slow on large ramps
            return set()  # HACK: makes this work for now
            # FIXME: please do

        upper2 = sorted(list(self.upper), key=lambda x: x.distance_to(self.bottom_center), reverse=True)
        while len(upper2) > 2:
            upper2.pop()
        return set(upper2)

    @property
    def top_center(self) -> Point2:
        pos = Point2(
            (sum([p.x for p in self.upper]) / len(self.upper), sum([p.y for p in self.upper]) / len(self.upper))
        )
        return pos

    @property
    def lower(self) -> Set[Point2]:
        min_height = min([self.height_at(p) for p in self._points])
        return {p for p in self._points if self.height_at(p) == min_height}

    @property
    def bottom_center(self) -> Point2:
        pos = Point2(
            (sum([p.x for p in self.lower]) / len(self.lower), sum([p.y for p in self.lower]) / len(self.lower))
        )
        return pos

    @property
    def barracks_in_middle(self) -> Point2:
        """ Barracks position in the middle of the 2 depots """
        if len(self.upper2_for_ramp_wall) == 2:
            points = self.upper2_for_ramp_wall
            p1 = points.pop().offset((self.x_offset, self.y_offset))
            p2 = points.pop().offset((self.x_offset, self.y_offset))
            # Offset from top point to barracks center is (2, 1)
            intersects = p1.circle_intersection(p2, 5 ** 0.5)
            anyLowerPoint = next(iter(self.lower))
            return max(intersects, key=lambda p: p.distance_to(anyLowerPoint))
        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @property
    def depot_in_middle(self) -> Point2:
        """ Depot in the middle of the 3 depots """
        if len(self.upper2_for_ramp_wall) == 2:
            points = self.upper2_for_ramp_wall
            p1 = points.pop().offset((self.x_offset, self.y_offset))  # still an error with pixelmap?
            p2 = points.pop().offset((self.x_offset, self.y_offset))
            # Offset from top point to depot center is (1.5, 0.5)
            intersects = p1.circle_intersection(p2, 2.5 ** 0.5)
            anyLowerPoint = next(iter(self.lower))
            return max(intersects, key=lambda p: p.distance_to(anyLowerPoint))
        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @property
    def corner_depots(self) -> Set[Point2]:
        """ Finds the 2 depot positions on the outside """
        if len(self.upper2_for_ramp_wall) == 2:
            points = self.upper2_for_ramp_wall
            p1 = points.pop().offset((self.x_offset, self.y_offset))  # still an error with pixelmap?
            p2 = points.pop().offset((self.x_offset, self.y_offset))
            center = p1.towards(p2, p1.distance_to(p2) / 2)
            depotPosition = self.depot_in_middle
            # Offset from middle depot to corner depots is (2, 1)
            intersects = center.circle_intersection(depotPosition, 5 ** 0.5)
            return intersects
        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @property
    def barracks_can_fit_addon(self) -> bool:
        """ Test if a barracks can fit an addon at natural ramp """
        # https://i.imgur.com/4b2cXHZ.png
        if len(self.upper2_for_ramp_wall) == 2:
            return self.barracks_in_middle.x + 1 > max(self.corner_depots, key=lambda depot: depot.x).x
        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")

    @property
    def barracks_correct_placement(self) -> Point2:
        """ Corrected placement so that an addon can fit """
        if len(self.upper2_for_ramp_wall) == 2:
            if self.barracks_can_fit_addon:
                return self.barracks_in_middle
            else:
                return self.barracks_in_middle.offset((-2, 0))
        raise Exception("Not implemented. Trying to access a ramp that has a wrong amount of upper points.")


class GameInfo:
    def __init__(self, proto):
        # TODO: this might require an update during the game because placement grid and
        # playable grid are greyed out on minerals, start locations and ramps (debris)
        # but we do not want to call information in the fog of war
        self._proto = proto
        self.players: List[Player] = [Player.from_proto(p) for p in self._proto.player_info]
        self.map_name: str = self._proto.map_name
        self.local_map_path: str = self._proto.local_map_path
        self.map_size: Size = Size.from_proto(self._proto.start_raw.map_size)
        self.pathing_grid: PixelMap = PixelMap(self._proto.start_raw.pathing_grid)
        self.terrain_height: PixelMap = PixelMap(self._proto.start_raw.terrain_height)
        self.placement_grid: PixelMap = PixelMap(self._proto.start_raw.placement_grid)
        self.playable_area = Rect.from_proto(self._proto.start_raw.playable_area)
        self.map_ramps: List[Ramp] = None  # Filled later by BotAI._prepare_first_step
        self.player_races: Dict[int, "Race"] = {
            p.player_id: p.race_actual or p.race_requested for p in self._proto.player_info
        }
        self.start_locations: List[Point2] = [Point2.from_proto(sl) for sl in self._proto.start_raw.start_locations]
        self.player_start_location: Point2 = None  # Filled later by BotAI._prepare_first_step

    @property
    def map_center(self) -> Point2:
        return self.playable_area.center

    def _find_ramps(self) -> List[Ramp]:
        """Calculate (self.pathing_grid - self.placement_grid) (for sets) and then find ramps by comparing heights."""
        rampDict = {
            Point2((x, y)): self.pathing_grid[(x, y)] == 0 and self.placement_grid[(x, y)] == 0
            for x in range(self.pathing_grid.width)
            for y in range(self.pathing_grid.height)
        }

        rampPoints = {p for p in rampDict if rampDict[p]}  # filter only points part of ramp
        rampGroups = self._find_groups(rampPoints)
        return [Ramp(group, self) for group in rampGroups]

    def _find_groups(
        self, points: Set[Point2], minimum_points_per_group: int = 8, max_distance_between_points: int = 2
    ) -> List[Set[Point2]]:
        """ From a set/list of points, this function will try to group points together """
        """ Paint clusters of points in rectangular map using flood fill algorithm. """
        NOT_INTERESTED = -2
        NOT_COLORED_YET = -1
        currentColor: int = NOT_COLORED_YET
        picture: List[List[int]] = [
            [NOT_INTERESTED for j in range(self.pathing_grid.width)] for i in range(self.pathing_grid.height)
        ]

        def paint(pt: Point2) -> None:
            picture[pt.y][pt.x] = currentColor

        nearby: Set[Point2] = set()
        for dx in range(-max_distance_between_points, max_distance_between_points + 1):
            for dy in range(-max_distance_between_points, max_distance_between_points + 1):
                if abs(dx) + abs(dy) <= max_distance_between_points:
                    nearby.add(Point2((dx, dy)))

        for point in points:
            paint(point)

        remaining: Set[Point2] = set(points)
        queue: Deque[Point2] = deque()
        foundGroups: List[Set[Point2]] = []
        while remaining:
            currentGroup: Set[Point2] = set()
            if not queue:
                currentColor += 1
                start = remaining.pop()
                paint(start)
                queue.append(start)
                currentGroup.add(start)
            while queue:
                base: Point2 = queue.popleft()
                for offset in nearby:
                    px, py = base.x + offset.x, base.y + offset.y
                    if px < 0 or py < 0 or px >= self.pathing_grid.width or py >= self.pathing_grid.height:
                        continue
                    if picture[py][px] != NOT_COLORED_YET:
                        continue
                    point: Point2 = Point2((px, py))
                    remaining.remove(point)
                    paint(point)
                    queue.append(point)
                    currentGroup.add(point)
            if len(currentGroup) >= minimum_points_per_group:
                foundGroups.append(currentGroup)

        """ Returns groups of points as list
        [{p1, p2, p3}, {p4, p5, p6, p7, p8}]
        """
        return foundGroups



from functools import wraps


def cache_forever(f):
    f.cache = {}

    @wraps(f)
    def inner(*args):
        if args not in f.cache:
            f.cache[args] = f(*args)
        return f.cache[args]

    return inner


def property_cache_forever(f):
    f.cached = None

    @wraps(f)
    def inner(self):
        if f.cached is None:
            f.cached = f(self)
        return f.cached

    return property(inner)


def property_cache_once_per_frame(f):
    """ This decorator caches the return value for one game loop, then clears it if it is accessed in a different game loop
    Only works on properties of the bot object because it requires access to self.state.game_loop """
    f.frame = -1
    f.cache = None

    @wraps(f)
    def inner(self):
        if f.frame != self.state.game_loop:
            f.frame = self.state.game_loop
            f.cache = None
        if f.cache is None:
            f.cache = f(self)
        return f.cache

    return property(inner)


def property_immutable_cache(f):
    @wraps(f)
    def inner(self):
        if f.__name__ not in self.cache:
            self.cache[f.__name__] = f(self)
        return self.cache[f.__name__]

    return property(inner)


def property_mutable_cache(f):
    @wraps(f)
    def inner(self):
        if f.__name__ not in self.cache:
            self.cache[f.__name__] = f(self)
        return self.cache[f.__name__].copy()

    return property(inner)



from .ids.ability_id import *
from .ids.buff_id import *
from .ids.effect_id import *
from .ids.unit_typeid import *
from .ids.upgrade_id import *

mineral_ids = {
    RICHMINERALFIELD.value,
    RICHMINERALFIELD750.value,
    MINERALFIELD.value,
    MINERALFIELD750.value,
    LABMINERALFIELD.value,
    LABMINERALFIELD750.value,
    PURIFIERRICHMINERALFIELD.value,
    PURIFIERRICHMINERALFIELD750.value,
    PURIFIERMINERALFIELD.value,
    PURIFIERMINERALFIELD750.value,
    BATTLESTATIONMINERALFIELD.value,
    BATTLESTATIONMINERALFIELD750.value,
}
geyser_ids = {
    VESPENEGEYSER.value,
    SPACEPLATFORMGEYSER.value,
    RICHVESPENEGEYSER.value,
    PROTOSSVESPENEGEYSER.value,
    PURIFIERVESPENEGEYSER.value,
    SHAKURASVESPENEGEYSER.value,
}



from setuptools import setup, find_packages

from pipenv.project import Project
from pipenv.utils import convert_deps_to_pip

pfile = Project(chdir=False).parsed_pipfile
requirements = convert_deps_to_pip(pfile['packages'], r=False)
test_requirements = convert_deps_to_pip(pfile['dev-packages'], r=False)

setup(
    name = "sc2-env",
    packages = find_packages(),
    version = "0.11.1",
    description = "A StarCraft II bot gym env library over python-sc2",
    license="MIT",
    author = "Damien GILLES",
    author_email = "damien.gilles.pro@gmail.com",
    url = "https://github.com/gillesdami/python-sc2",
    keywords = ["StarCraft", "StarCraft 2", "StarCraft II", "AI", "Bot"],
    setup_requires=["pipenv"],
    install_requires=requirements,
    classifiers = [
        "Development Status :: 3 - Alpha",

        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Games/Entertainment",
        "Topic :: Games/Entertainment :: Real Time Strategy",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",

        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ]
)



import asyncio
import async_timeout
import time

import logging
logger = logging.getLogger(__name__)

from .sc2process import SC2Process
from .portconfig import Portconfig
from .client import Client
from .player import Human, Bot
from .data import Race, Difficulty, Result, ActionResult, CreateGameError
from .game_state import GameState
from .protocol import ConnectionAlreadyClosed, ProtocolError

class SlidingTimeWindow:
    def __init__(self, size: int):
        assert size > 0

        self.window_size = size
        self.window = []

    def push(self, value: float):
        self.window = (self.window + [value])[-self.window_size:]

    def clear(self):
        self.window = []

    @property
    def sum(self) -> float:
        return sum(self.window)

    @property
    def available(self) -> float:
        return sum(self.window[1:])

    @property
    def available_fmt(self) -> float:
        return ",".join(f"{w:.2f}" for w in self.window[1:])


async def _play_game_human(client, player_id, realtime, game_time_limit):
    while True:
        state = await client.observation()
        if client._game_result:
            return client._game_result[player_id]

        if game_time_limit and (state.observation.observation.game_loop * 0.725 * (1 / 16)) > game_time_limit:
            print(state.observation.game_loop, state.observation.game_loop * 0.14)
            return Result.Tie

        if not realtime:
            await client.step()

async def _play_game_ai(client, player_id, ai, realtime, step_time_limit, game_time_limit):
    if realtime:
        assert step_time_limit is None

    # step_time_limit works like this:
    # * If None, then step time is not limited
    # * If given integer or float, the bot will simpy resign if any step takes longer than that
    # * Otherwise step_time_limit must be an object, with following settings:
    #
    # Key         | Value      | Description
    # ------------|------------|-------------
    # penalty     | None       | No penalty, the bot can continue on next step
    # penalty     | N: int     | Cooldown penalty, BotAI.on_step will not be called for N steps
    # penalty     | "resign"   | Bot resigns when going over time limit
    # time_limit  | int/float  | Time limit for a single step
    # window_size | N: int     | The time limit will be used for last N steps, instad of 1
    #
    # Cooldown is a harsh penalty. The both loses the ability to act, but even worse,
    # the observation data from skipped steps is also lost. It's like falling asleep in
    # a middle of the game.
    time_penalty_cooldown = 0
    if step_time_limit is None:
        time_limit = None
        time_window = None
        time_penalty = None
    elif isinstance(step_time_limit, (int, float)):
        time_limit = float(step_time_limit)
        time_window = SlidingTimeWindow(1)
        time_penalty = "resign"
    else:
        assert isinstance(step_time_limit, dict)
        time_penalty = step_time_limit.get("penalty", None)
        time_window = SlidingTimeWindow(int(step_time_limit.get("window_size", 1)))
        time_limit = float(step_time_limit.get("time_limit", None))


    game_data = await client.get_game_data()
    game_info = await client.get_game_info()

    ai._prepare_start(client, player_id, game_info, game_data)
    try:
        ai.on_start()
    except Exception as e:
        logger.exception(f"AI on_start threw an error")
        logger.error(f"resigning due to previous error")
        ai.on_end(Result.Defeat)
        return Result.Defeat

    iteration = 0
    while True:
        state = await client.observation()
        logger.debug(f"Score: {state.observation.observation.score.score}")

        if client._game_result:
            ai.on_end(client._game_result[player_id])
            return client._game_result[player_id]

        gs = GameState(state.observation, game_data)

        if game_time_limit and (gs.game_loop * 0.725 * (1 / 16)) > game_time_limit:
            ai.on_end(Result.Tie)
            return Result.Tie

        ai._prepare_step(gs)

        if iteration == 0:
            ai._prepare_first_step()

        logger.debug(f"Running AI step, it={iteration} {gs.game_loop * 0.725 * (1 / 16):.2f}s")

        try:
            await ai.issue_events()
            if realtime:
                await ai.on_step(iteration)
            else:
                if time_penalty_cooldown > 0:
                    time_penalty_cooldown -= 1
                    logger.warning(f"Running AI step: penalty cooldown: {time_penalty_cooldown}")
                    iteration -= 1 # Do not increment the iteration on this round
                elif time_limit is None:
                    await ai.on_step(iteration)
                else:
                    out_of_budget = False
                    budget = time_limit - time_window.available

                    # Tell the bot how much time it has left attribute
                    ai.time_budget_available = budget

                    if budget < 0:
                        logger.warning(f"Running AI step: out of budget before step")
                        step_time = 0.0
                        out_of_budget = True
                    else:
                        step_start = time.monotonic()
                        try:
                            async with async_timeout.timeout(budget):
                                await ai.on_step(iteration)
                        except asyncio.TimeoutError:
                            step_time = time.monotonic() - step_start
                            logger.warning(
                                f"Running AI step: out of budget; " +
                                f"budget={budget:.2f}, steptime={step_time:.2f}, " +
                                f"window={time_window.available_fmt}"
                            )
                            out_of_budget = True
                        step_time = time.monotonic() - step_start

                    time_window.push(step_time)

                    if out_of_budget and time_penalty != None:
                        if time_penalty == "resign":
                            raise RuntimeError("Out of time")
                        else:
                            time_penalty_cooldown = int(time_penalty)
                            time_window.clear()
        except Exception as e:
            if isinstance(e, ProtocolError) and e.is_game_over_error:
                if realtime:
                    return None
                result = client._game_result[player_id]
                if result is None:
                    log.error("Game over, but no results gathered")
                    raise
                return result

            # NOTE: this message is caught by pytest suite
            logger.exception(f"AI step threw an error") # DO NOT EDIT!
            logger.error(f"resigning due to previous error")
            ai.on_end(Result.Defeat)
            return Result.Defeat

        logger.debug(f"Running AI step: done")

        if not realtime:
            if not client.in_game:  # Client left (resigned) the game
                ai.on_end(client._game_result[player_id])
                return client._game_result[player_id]

            await client.step()

        iteration += 1

async def _play_game(player, client, realtime, portconfig, step_time_limit=None, game_time_limit=None, rgb_render_config=None):
    assert isinstance(realtime, bool), repr(realtime)

    player_id = await client.join_game(player.name, player.race, portconfig=portconfig, rgb_render_config=rgb_render_config)
    logging.info(f"Player id: {player_id} ({player.name})")

    if isinstance(player, Human):
        result = await _play_game_human(client, player_id, realtime, game_time_limit)
    else:
        result = await _play_game_ai(client, player_id, player.ai, realtime, step_time_limit, game_time_limit)

    logging.info(f"Result for player id: {player_id}: {result}")
    return result

async def _setup_host_game(server, map_settings, players, realtime, random_seed=None):
    r = await server.create_game(map_settings, players, realtime, random_seed)
    if r.create_game.HasField("error"):
        err = f"Could not create game: {CreateGameError(r.create_game.error)}"
        if r.create_game.HasField("error_details"):
            err += f": {r.create_game.error_details}"
        logger.critical(err)
        raise RuntimeError(err)

    return Client(server._ws)


async def _host_game(map_settings, players, realtime, portconfig=None, save_replay_as=None, step_time_limit=None,
                     game_time_limit=None, rgb_render_config=None, random_seed=None):
    assert len(players) > 0, "Can't create a game without players"

    assert any(isinstance(p, (Human, Bot)) for p in players)

    async with SC2Process(render=rgb_render_config is not None) as server:
        await server.ping()

        client = await _setup_host_game(server, map_settings, players, realtime, random_seed)

        try:
            result = await _play_game(players[0], client, realtime, portconfig, step_time_limit, game_time_limit, rgb_render_config)
            if save_replay_as is not None:
                await client.save_replay(save_replay_as)
            await client.leave()
            await client.quit()
        except ConnectionAlreadyClosed:
            logging.error(f"Connection was closed before the game ended")
            return None

        return result

async def _host_game_aiter(map_settings, players, realtime, portconfig=None, save_replay_as=None, step_time_limit=None, game_time_limit=None):
    assert players, "Can't create a game without players"

    assert any(isinstance(p, (Human, Bot)) for p in players)

    async with SC2Process() as server:
        while True:
            await server.ping()

            client = await _setup_host_game(server, map_settings, players, realtime)

            try:
                result = await _play_game(players[0], client, realtime, portconfig, step_time_limit, game_time_limit)

                if save_replay_as is not None:
                    await client.save_replay(save_replay_as)
                await client.leave()
            except ConnectionAlreadyClosed:
                logging.error(f"Connection was closed before the game ended")
                return

            new_players = yield result
            if new_players is not None:
                players = new_players

def _host_game_iter(*args, **kwargs):
    game = _host_game_aiter(*args, **kwargs)
    new_playerconfig = None
    while True:
        new_playerconfig = yield asyncio.get_event_loop().run_until_complete(game.asend(new_playerconfig))


async def _join_game(players, realtime, portconfig, save_replay_as=None, step_time_limit=None, game_time_limit=None):
    async with SC2Process() as server:
        await server.ping()

        client = Client(server._ws)

        try:
            result = await _play_game(players[1], client, realtime, portconfig, step_time_limit, game_time_limit)
            if save_replay_as is not None:
                await client.save_replay(save_replay_as)
            await client.leave()
            await client.quit()
        except ConnectionAlreadyClosed:
            logging.error(f"Connection was closed before the game ended")
            return None

        return result

def run_game(map_settings, players, **kwargs):
    if sum(isinstance(p, (Human, Bot)) for p in players) > 1:
        host_only_args = ["save_replay_as", "rgb_render_config", "random_seed"]
        join_kwargs = {k: v for k, v in kwargs.items() if k not in host_only_args}

        portconfig = Portconfig()
        result = asyncio.get_event_loop().run_until_complete(asyncio.gather(
            _host_game(map_settings, players, **kwargs, portconfig=portconfig),
            _join_game(players, **join_kwargs, portconfig=portconfig)
        ))
    else:
        result = asyncio.get_event_loop().run_until_complete(
            _host_game(map_settings, players, **kwargs)
        )
    return result
